from scipy.integrate import solve_ivp
import numpy as np
import matplotlib.pyplot as plt
import math
import scipy as sp
from scipy.interpolate import CubicSpline
from scipy.optimize import fsolve
from scipy.interpolate import splev, splrep 
import time
import os
from datetime import datetime


#interpolation RR with cubic splain
def interpolateCubic(SourceSignal,Fsamp):
    t=np.sum(SourceSignal)
    x = np.linspace(0 , t , len(SourceSignal))
    cs = CubicSpline(x, SourceSignal)
    x2 = np.linspace(0 , t , t*Fsamp)
    y2 = cs(x2)
    return(y2)

##filtrating witn band pass filter (minFriq,maxFriq - frequency-limits of the useful signal)
def band_pass_filter (array1, minFriq, maxFriq, Fsamp):
    spectrum = np.fft.rfft(array1) ##ищем спектр сигнала
    ##Фильтруем спектр в указанной полосе
    fn = 1 / (2 / Fsamp)
    friqArray = np.linspace(0, fn, len(spectrum))
    minIndex=0
    maxIndex=0
    for j in range (1, len(friqArray)):
        if ((friqArray[j]==minFriq)or((friqArray[j-1]<minFriq)and(friqArray[j]>minFriq))):
            minIndex=j
        if ((friqArray[j]==maxFriq)or((friqArray[j-1]<maxFriq)and(friqArray[j]>maxFriq))):
            maxIndex=j
    spectrum[0:minIndex]=0
    spectrum[maxIndex:len(friqArray)]=0

    FilterSignal=np.fft.irfft(spectrum, len(array1))
    return FilterSignal


def resamp(array, Fsamp, F_reSamp): 
    t=len(array)/Fsamp

    ResultSignal=np.zeros(int(t*F_reSamp))
    l=0
    for j in range (0,len(ResultSignal)):
        ResultSignal[j]=array[l] 
        l=l+Fsamp//F_reSamp
    return ResultSignal

##get unwrap phases using Hilbert transform
def GetHilbert (array):
    Phase1=np.zeros(len(array))
    spectrum=np.fft.rfft(array)
    addArray=np.fft.irfft(spectrum*1j)
    if (len(addArray)<len(array)):
        array=array[0:len(addArray)]
    else:
        addArray=addArray[0:len(array)]
    
    Ph=-1
    JumpsCount=0
    for i in range(0,len(array)):
        PhS=-np.arctan2(addArray[i],array[i])/np.pi

        if (Ph==-1):
            Phold=PhS

        if ((Phold<0) and (PhS>0)):
            JumpsCount+=1
        Phold=PhS

        if (PhS<0):
            PhS=2+PhS
        if (Ph==-1):
            Ph=0
            Ph1=PhS
        else:
            Ph=PhS-Ph1+2*JumpsCount
        Phase1[i]=Ph
    return (Phase1)


def PreparePolyFit(Win):
    iSum=0
    Ni2Sum=0
    iSum2=0
    L2=Win//2

    for i in range (-L2,L2+1):
        iSum+=i
        Ni2Sum+=i*i
    Ni2Sum=Ni2Sum*Win
    iSum2=iSum*iSum
    Ni2SumiSum2=Ni2Sum+iSum2
    return Ni2SumiSum2

def FitPoly(difP,Fsamp,Win,N):
    difP2=[]
    for i in range (0,len(difP)):
        difP2.append(difP[i])
    difP2.append(0)
    L2=Win//2
    start=N-L2
    end=N+L2
    a=0
    b=0
    YiSum=0
    iYiSum=0

    if (N>=start) and (N<=end):
        for I in range(N-L2,N+L2+1):
            YiSum=YiSum+difP2[I]
            iYiSum=iYiSum+(I-N)*difP2[I]
            
    Ni2SumiSum2=PreparePolyFit(Win) 
    a=Win*iYiSum/Ni2SumiSum2
    b=YiSum/Win;
    return(a,b) 

##helps to avoid jumps in phase difference    
def CorrectDifPhases(difP):
    corDif=np.zeros(len(difP))
    y=[]
    for i in range (0,len(difP)-1):
        y.append(np.abs(difP[i+1]-difP[i]))
        
    for i in range (0,len(difP)-1):
        corDif[i]=corDif[i]+difP[i]
        if (np.abs(difP[i+1]-difP[i])>10*np.median(y)):
            if (difP[i+1]>difP[i]):
                for j in range(i+1,len(difP)):
                    corDif[j]=corDif[j]-((difP[i+1]-difP[i])+np.median(y))
            else:
                for j in range(i+1,len(difP)):
                    corDif[j]=corDif[j]+((difP[i]-difP[i+1])-np.median(y))
    corDif[len(difP)-1]=corDif[len(difP)-2]
    return corDif

##searchs epochs of synchronization    
def ShelfSearcher(difP,Win,aLim,aLimShelf,Lmin): 

    Step=1
    N1=Win // 2
##    print(N1)
    N2=len(difP)-(Win // 2)-1
##    print(N2)
    N=N1
##    OldLen=0
    Len=0
    a=0
    b=0
    Shelfs_Len=[]
    Shelfs_X=[]
    Shelfs_Y=[]
    while (N<=N2):
        a,b=FitPoly(difP,Fsamp,Win,N)
        while(np.abs(a)<aLim) and (N<=N2):
            Len+=1
            N=N+Step
            a,b=FitPoly(difP,Fsamp,Win,N)
        if (Len>Lmin):
            OldLen=Win
            Win=Len
            a,b=FitPoly(difP,Fsamp,Win,(N-Len//2-1))
            if (np.abs(a)<aLimShelf):
                Shelfs_Len.append(Len) ##length of synchron epoch
                Shelfs_X.append(N-Len)
                Shelfs_Y.append(b)
            Win=OldLen
        Len=0
        N=N+Step
    return Shelfs_Len,Shelfs_X,Shelfs_Y

##AAFT surrogates
def AAFT (x1):
    #search spectra of source signal 
    x = np.abs(np.fft.rfft(x1))
    f=x*np.exp(1j*np.random.uniform(0,2*np.pi,len(x)))
    ##recover new signal from spectra
    f1=np.fft.irfft(f)
    return f1

def Rights (path):

    rights=[]
    f = open(path,"r")
    s = f.readline()
    while s:
        rights.append(s)
        s = f.readline()

    for i in range (0,len(rights)):
        if (rights[i][0]=='1'):
            if (rights[i].find('#')!=-1):
                Fsamp=rights[i][rights[i].find('=')+1:rights[i].find('#')]
            else:
                Fsamp=rights[i][rights[i].find('=')+1:len(rights[i])-1]
            Fsamp=Fsamp.strip(' ')

        if (rights[i][0]=='2'):
            if (rights[i].find('#')!=-1):
                F_reSamp=rights[i][rights[i].find('=')+1:rights[i].find('#')]
            else:
                F_reSamp=rights[i][rights[i].find('=')+1:len(rights[i])-1]
            F_reSamp=F_reSamp.strip(' ')

        if (rights[i][0]=='3'):
            if (rights[i].find('#')!=-1):
                pathIn1=rights[i][rights[i].find('=')+1:rights[i].find('#')]
            else:
                pathIn1=rights[i][rights[i].find('=')+1:len(rights[i])-1]
            pathIn1=pathIn1.strip(' ')

        if (rights[i][0]=='4'):
            if (rights[i].find('#')!=-1):
                pathIn2=rights[i][rights[i].find('=')+1:rights[i].find('#')]
            else:
                pathIn2=rights[i][rights[i].find('=')+1:len(rights[i])-1]
            pathIn2=pathIn2.strip(' ')

    rights=[]
    rights.append(Fsamp)
    rights.append(F_reSamp)
    rights.append(pathIn1)
    rights.append(pathIn2)

    return(rights)

if __name__=="__main__":    
    rights=Rights('Data\Config.txt')    
    Fsamp=int(rights[0])
    F_reSamp=int(rights[1])
    pathIN1=rights[2]
    pathIN2=rights[3]
    pathOUT='Results'
    n1=0
    n2=300
    nSurrogates=2
    
    if not os.path.isdir(pathOUT):
        os.mkdir(pathOUT)
        
    pathOUT=pathOUT+'\\'+str(datetime.strftime(datetime.today(),"%d_%m_%y %H_%M_%S"))

    os.makedirs(pathOUT)
    
    with open(pathOUT+'\\report.txt', 'w') as fileRep:
        fileRep.write('Data:'+'\n')
        fileRep.write('\n')


        Win=22 
        aLim=0.004 
        aLimShelf=0.01 
        Lmin=10 
        minFriq=0.06 
        maxFriq=0.14 

        

        print('-> loading RR intervals')
        fileRep.write('File RR-intervals'+'\n')
        fileRep.write(os.path.abspath(pathIN1)+'\n')
        signalRR=np.loadtxt(pathIN1,skiprows=n1)
        fileRep.write('Signal length: '+str(np.sum(signalRR))+' с'+'\n')
        fileRep.write('\n')
        
        print('-> loading BP signal')
        fileRep.write('File BP'+'\n')
        fileRep.write(os.path.abspath(pathIN2)+'\n')
        signalPPG=np.loadtxt(pathIN2,skiprows=n1)
        fileRep.write('Signal length: '+str(len(signalPPG)/Fsamp)+' с'+'\n')
        fileRep.write('\n')

        fileRep.write('Fsamp                 = '+str(Fsamp)+' Гц \n')
        fileRep.write('F_reSamp              = '+str(F_reSamp)+' Гц \n')
        fileRep.write('\n')

        print('-> interpolating RR signal')
        intRR=interpolateCubic(signalRR,Fsamp)


        if(len(intRR)<len(signalPPG)): 
            signalPPG=signalPPG[0:len(intRR)]
        else:
            intRR=intRR[0:len(signalPPG)]

        if ((n2*Fsamp<len(intRR))and(n2*Fsamp<len(signalPPG))and(n2!=0)):
            intRR=intRR[0:(n2*Fsamp)]
            signalPPG=signalPPG[0:(n2*Fsamp)]
        else:
            print('Signal length is less than specified in the settings!')
            
        print('-> filtrating RR signal')
        filtArrayRR=band_pass_filter(intRR,minFriq,maxFriq,Fsamp)
        arrayRR=filtArrayRR/np.max(np.abs(filtArrayRR))


        print('-> filtrating BP signal')
        filtArrayPPG=band_pass_filter(signalPPG,minFriq,maxFriq,Fsamp)
        arrayPPG=filtArrayPPG/np.max(np.abs(filtArrayPPG))
        fileRep.write('To exclude edge filtering effects, the section of realizations (10-290 s) was analyzed '+'\n')
        fileRep.write('\n')
    
        print('-> resempling BP signal')
        if (F_reSamp<Fsamp):
            arrayPPG=resamp(arrayPPG,Fsamp,F_reSamp)
            arrayRR=resamp(arrayRR,Fsamp,F_reSamp)


        f = open(pathOUT+'\\filtred_RR.txt', 'w')
        for j in range(0,len(arrayRR)):
            f.write(str(arrayRR[j])+'\n')
        f.close()    
        f = open(pathOUT+'\\filtred_BP.txt', 'w')
        for j in range(0,len(arrayPPG)):
            f.write(str(arrayPPG[j])+'\n')
        f.close()

        start_time = time.time() ## start calculating time

        print('-> calculating phase for RR signal')
        phase1=GetHilbert(arrayRR[50:len(arrayRR)-50])
        f = open(pathOUT+'\\phase_RR.txt', 'w')
        for j in range(0,len(phase1)):
            f.write(str(phase1[j])+'\n')
        f.close()

        print('-> calculating phase for BP signal')
        phase2=GetHilbert(arrayPPG[50:len(arrayPPG)-50])
        f = open(pathOUT+'\\phase_BP.txt', 'w')
        for j in range(0,len(phase2)):
            f.write(str(phase2[j])+'\n')
        f.close()


        print('-> calculating phase difference')
        dif=phase1-phase2
        CorDif=np.abs(CorrectDifPhases(dif))
        with open(pathOUT+'\\phase difference.txt', 'w') as f:
            for j in range(0,len(CorDif)):
                f.write(str(CorDif[j])+'\n')

        print('-> search syncronization epochs')
        Len,X,Y=ShelfSearcher(CorDif,Win,aLim,aLimShelf,Lmin)
        fileRep.write('Results: \n')
   
        print('-> calculating S,%')
        S=(np.sum(Len)/F_reSamp)/((len(CorDif)-Win)/F_reSamp)*100
        print('S = '+'%.2f' % S + ' %')

        time1=time.time() - start_time ## time for whole algorizm without AAFT
        strTime='%.0f' % time1
        print('-> The program has already been running for '+strTime+ ' seconds')
        if (nSurrogates>0):
            time1=time1*nSurrogates
            strTime='%.0f' % time1

        
            ##generation AAFT surrogates (numbers = nSurrogates)
            print('-> start calculating AAFT surrogates - it will take about '+strTime+ ' seconds')
            print('-> 0 % done')
            artificialS=[]
            proc_time=int(100/nSurrogates)
            k=0

            for j in range (0,nSurrogates):
                artificialArray1=AAFT(arrayRR)
                artificialArray2=AAFT(arrayPPG)
                Aphase1=GetHilbert(artificialArray1[50:len(artificialArray1)-50])
                Aphase2=GetHilbert(artificialArray2[50:len(artificialArray2)-50])
                Adif=np.abs(Aphase1-Aphase2)
                ACorDif=CorrectDifPhases(Adif)
                ALen,AX,AY=ShelfSearcher(ACorDif,Win,aLim,aLimShelf,Lmin)
                artificialS.append((np.sum(ALen)/F_reSamp)/((len(ACorDif)-Win)/F_reSamp)*100)
                k+=proc_time
                print('-> '+str(k)+' % done')            

            p=1.00
            artificialS.sort()
            for j in range(0,len(artificialS)-1):
                if (S>=artificialS[j]):
                    p-=(1/len(artificialS))
            print('p < '+ '%.3f' % p)
            fileRep.write('S = '+ '%.1f' % S + '% ( p < '+ '%.3f' % p + ' )')
            fileRep.write('\n')

        else:
            fileRep.write('S = '+ '%.1f' % S + '% ')
            fileRep.write('\n')
        fileRep.write('\n') 
        fileRep.write('Synchronization epochs (relative to the beginning of the source file), seconds: \n')
        fileRep.write('start end length \n')
        for i in range(0,len(X)):
                      fileRep.write(str(X[i]/F_reSamp)+' '+str((X[i]+Len[i])/F_reSamp)+' '+str(Len[i]/F_reSamp)+'\n')
        fileRep.write('\n')

        time2=time.time() - start_time ## time for whole algorizm with AAFT
        strTime='%.0f' % time2
        fileRep.write('Computation time: '+strTime+ ' seconds \n')
                    
        ##open plots
        fig1=plt.figure(facecolor='pink',figsize=(10,7))
        fig1.canvas.set_window_title('Results')
        timeArray1=np.linspace(0,len(CorDif)/F_reSamp,len(CorDif))
        plt.subplot(3, 1, 1)
        plt.plot(timeArray1,CorDif*np.pi,'k')
        for i in range(0,len(X)):
            plt.axvspan(X[i]/F_reSamp, (X[i]+Len[i]+1)/F_reSamp, color='y', alpha=0.4, lw=0)
        plt.xlabel('$\\it{t}$'+' (s)',fontsize=14)
        plt.ylabel('$ \mathit{\Delta} \phi$'+' (rad)',fontsize=14)
        plt.grid()
        plt.title('yellow areas - synchronization epochs',fontsize=14)
        plt.subplot(3, 1, 2)
        timeArray1=np.linspace(0,(len(intRR))/Fsamp-100/F_reSamp,len(intRR)-(100/F_reSamp)*Fsamp)
##        x = np.linspace(0 , t , len(SourceSignal)) 
##        plt.plot(timeArray1,SourceSignal[50:len(intRR)-50],'k')
        plt.plot(timeArray1,intRR[int((50/F_reSamp)*Fsamp):len(intRR)-int((50/F_reSamp)*Fsamp)],'r')
        plt.xlabel('$\\it{t}$'+' (s)',fontsize=14)
        plt.ylabel('RR-intervals '+' (s)',fontsize=14)
        plt.grid()
        plt.subplot(3, 1, 3)
        timeArray1=np.linspace(0,(len(signalPPG))/Fsamp-100/F_reSamp,len(signalPPG)-(100/F_reSamp)*Fsamp)
        plt.plot(timeArray1,signalPPG[int((50/F_reSamp)*Fsamp):len(signalPPG)-int((50/F_reSamp)*Fsamp)],'b')
        plt.xlabel('$\\it{t}$'+' (s)',fontsize=14)
        plt.ylabel('BP '+' (mm Hg)',fontsize=14)
        plt.grid()
        plt.subplots_adjust(hspace=0.5)
        fig1.savefig(pathOUT+'\\'+'phase difference.png', format='png', dpi=100)
        plt.show()
        
                

            

            

    



    
